#include "npc.h"

/* Each set is represented in terms of a struct set whose index is the first
   index of a variable (used in a formula), a couple of flags to indicate if
   true and false respectively are part of satisfying truth assignments, and
   pointers to the struct set indicating what assignments of successive
   variables make the set satisfiable in conjunction with true and false.
   I.E. the set consisting of one element which is true in the variable of
   index 1 and false in the variable of index 2, is represented by a struct
   set with an index of 1, SATISFY_TRUE set but not SATISFY_FALSE, and the
   pointer false pointing to NULL, and true pointing to a set with index 2,
   SATISFY_FALSE set but not SATISFY_TRUE, and both pointers pointing to NULL.

   The algorithm does not work with individual truth assignments, but rather
   with sets of assignments which are treated as a single unit.  So we deal
   with a set of truth assignments in polynomial time, instead of every single
   truth assignment, considered separately, in polynomial time each.  The
   algorithm develops linked lists which specify a certain number of
   variables, with a NULL pointer at the end of a specification, so that any
   assignment of variables not specified (for one specification) satisfies the
   formula.  The set of satisfying truth assignments is then built up, having
   the structure of a tree, a tree of the same complexity level as the parsing
   tree of the formula.

   Pseudocode is as follows:

   set complement(set s)
	Allocate and initialize a result variable, of the same index as the
	input set.
	If s is a NULL pointer, set the result to satisfy both true and
		false for any combination of subsequent variables.
	Else:
		If s is satisfied by true and any assignment of unspecified
			variables, then the result is satisfied by false and
			any assignment of unspecified variables.
		If s is satisfied by false and any assignment of unspecified
			variables, then the result is satisfied by false and
			any assignment of unspecified variables.
		If s is satisfied by true and only some assignments of
			subsequent variables, then s is satisfied by true and
			the complement of the set of assignments of subsequent
			variables.
		If s is satisfied by false and only some assignments of
			subsequent variables, then s is satisfied by true and
			the complement of the set of assignments of subsequent
			variables.

   set intersection(set s1, set s2)
	If one set does not specify any variables (i.e. is referenced by a
		null pointer), then the result is the other set.  If neither
		specify any variables, then the result doesn't specify any
		variables either.
	Otherwise:
		Find out which one has a lower index (first) and which a higher
			index (second), if the indices are different.
		Allocate and initialize a set for the result, with index equal
			to the index of the first set, and neither true nor
			false satisfying.
		If the first set is of strictly lower index than the second:
			If the first set is satisfied by true, the result is
				satisfied by true, by the intersection of the
				set of subsequent satisfying assignments for
				true in the first set, and the second set.
			If the first set is satisfied by false, the result is
				satisfied by false, by the intersection of the
				set of subsequent satisfying assignments for
				false in the first set, and the second set.
		Else:
			If both sets are satisfied by true, then the result is
				satisfied for true, by the intersection of the
				set of subsequent satisfying assignments for
				true in the first set, and the set of
				subsequent satisfying assignments for true in
				the second set.
			If both sets are satisfied by false, then the result is
				satisfied for false, by the intersection of the
				set of subsequent satisfying assignments for
				false in the first set, and the set of
				subsequent satisfying assignments for false in
				the second set.

   set satisfy(formula f)
	If f is a single variable, allocate, initialize, and return a set with
		index equal to that variable, satisfied by true and not by
		false, with both pointers NULL.
	If f is (!f1) for some formula f1, return negation(satisfy(f1)).
	If f is (f1^f2) for some formulas f1 and f2, return
		intersection(satisfy(f1),satisfy(f2)).

   End of pseudocode.  */

struct set *complement(struct set *s)
	{
	struct set *result;
	result = (struct set *) malloc(sizeof(struct set));
	result->index=s->index;
	result->flags=0;
	result->true=result->false=NULL;
	if (s==NULL)
		{
		result->flags=(result->flags|(SATISFY_TRUE|SATISFY_FALSE));
		}
	else
		{
		if (!(s->flags & SATISFY_TRUE))
			result->flags=(result->flags|SATISFY_TRUE);
       		if (!(s->flags & SATISFY_FALSE))
			result->flags=(result->flags|SATISFY_FALSE);
		if ((!(result->flags & SATISFY_TRUE)) && (!(s->true==NULL)))
			{
			result->true=complement(s->true);
			if (!(empty(result->true)))
				result->flags=(result->flags|SATISFY_TRUE);
			}
		if ((!(result->flags & SATISFY_FALSE)) && (!(s->false==NULL)))
			{
             		result->false=complement(s->false);
			if (!(empty(result->false)))
              			result->flags=(result->flags|SATISFY_FALSE);
			}
		}
	/* if (result->flags & SATISFY_TRUE)
		printf("True: %d\n", s->index);
	if (result->flags & SATISFY_FALSE)
		printf("False: %d\n", s->index); */
	return result;
	}

struct set *intersection(struct set *s1, struct set *s2)
	{
	struct set *result, *first, *second;
	/* A null pointer references the whole set.  (Somewhat
	   counterintuitive, but the data contained tells of restrictions, and
	   a null pointer indicates no restrictions, i.e. the whole set.)  We
	   first check if one or both pointers are null.  */
	if (s1==NULL)
		result=s2;
	else if (s2==NULL)
		result=s1;
	else
		{
		if (s1->index <= s2->index)
			{
			first = s1;
			second = s2;
			}
		else
			{
			first = s2;
			second = s1;
			}
		result = (struct set *) malloc(sizeof(struct set));
		result->index=first->index;
		result->flags=0;
		result->true=result->false=NULL;
		if (first->index < second->index)
			{
			if (first->flags&SATISFY_TRUE)
				{
				result->true=intersection(first->true, second);
				if (!(empty(result->true)))
					result->flags=(result->flags|SATISFY_TRUE);
				}
			if (first->flags&SATISFY_FALSE)
				{
				result->false=intersection(first->false, second);
				if (!(empty(result->false)))
					result->flags=(result->flags|SATISFY_FALSE);
				}
			}
		else
			{
			if ((first->flags & SATISFY_TRUE) && (second->flags &
				SATISFY_TRUE))
				{
				result->true=intersection(first->true, second->true);
				if (!(empty(result->true)))
					result->flags=(result->flags|SATISFY_TRUE);
				}
			if ((first->flags & SATISFY_FALSE) && (second->flags &
				SATISFY_FALSE))
				{
				result->false=intersection(first->false, second->false);
				if (!(empty(result->false)))
					result->flags=(result->flags|SATISFY_FALSE);
				}
			}
		}
	return result;
	}

#ifdef USE_OLD_PARSER

struct set *satisfy(struct formula *f)
	{
	struct formula *curr_f, *sub1, *sub1_tail, *sub2, *sub2_tail;
	struct set *result, *temp1, *temp2;
	long left_parentheses;
	left_parentheses=0;
	if ((f->character >= '0') && (f->character <= '9')) /* a variable */
		{
		result=(struct set *) malloc(sizeof(struct set));
		result->index=calc_index(f);
		result->flags=SATISFY_TRUE;
		result->true=result->false=NULL;
		}
	else /* negation or conjunction */
		{
		if (f->next->character == '!') /* negation */
			{
			sub1=(struct formula *) malloc(sizeof(struct formula));
			sub1->character=f->next->next->character;
			sub1->next=NULL;
			sub1_tail=sub1;
			if (!(f->next->next->next->character==')'))
				/* If this condition is not met, we have the
				   negation of a variable of single digit
				   index, so sub1 is complete */
				{
                                for(curr_f=f->next->next;(!(left_parentheses==-1));)
                                        {
					curr_f=curr_f->next;
                                        if (curr_f->character=='(')
                                                ++left_parentheses;
                                        if (curr_f->character==')')
                                                --left_parentheses;
                                        append(sub1_tail, curr_f->character);
                                        sub1_tail=sub1_tail->next;
                                        }
				}
			temp1=satisfy(sub1);
			/* My garbage collection was causing problems by
			   pre-mature deletia; as the whole algorithm takes up
			   very little memory, I am leaving it out, to deal
			   with later.  */
			/* free_formula(sub1); */
			result=complement(temp1);
			/* free_set(temp1); */
			}
		else /* conjunction */
			{
			sub1=(struct formula *) malloc(sizeof(struct formula));
			sub1->character=f->next->character;
			left_parentheses=0;
			if ((sub1->character)=='(')
				++left_parentheses;
			sub1->next=NULL;
			sub1_tail=sub1;
			if (f->next->next->character=='^')
				/* single digit variable; we have read it
				into sub1 and now need only point the pointer
				to the '^' to be in the same place as the loop
				leaves it.  */
				curr_f=f->next->next;
			else /* read everything until parentheses are balanced
				and the next character is a '^' */
                        	for(curr_f=f->next;(!((left_parentheses==0) && (curr_f->character=='^')));)
                                	{
					curr_f=curr_f->next;
                                	if (curr_f->character=='(')
                                        	++left_parentheses;
                                	if (curr_f->character==')')
                                        	--left_parentheses;
					if (!((curr_f->character=='^') &&
						(left_parentheses==0)))
						{
                                		append(sub1_tail, curr_f->character);
                                		sub1_tail=sub1_tail->next;
						}
                                	}
			sub2=(struct formula *) malloc(sizeof(struct formula));
			sub2->character=curr_f->next->character;
			left_parentheses=-1;
			if (sub2->character=='(')
				++left_parentheses;
			sub2->next=NULL;
			sub2_tail=sub2;
			curr_f=curr_f->next;
			if (!(curr_f->next->character==')'))
				/* We are done if there's a single digit
				   variable; we only need to do more if it
				   isn't.  */
				/* Read in everything left but the last
				   character.  */
                        	for(;(!(curr_f->next==NULL));)
                                	{
					curr_f=curr_f->next;
                                	if (curr_f->character=='(')
                                        	++left_parentheses;
                                	if (curr_f->character==')')
                                        	--left_parentheses;
					if (!(left_parentheses==-1))
						{
                                		append(sub2_tail, curr_f->character);
                                		sub2_tail=sub2_tail->next;
						}
                                	}
			temp1=satisfy(sub1);
			/* free_formula(sub1); */
			temp2=satisfy(sub2);
			/* free_formula(sub2); */
			result=intersection(temp1, temp2);
			/* free_set(temp1); */
			/* free_set(temp2); */
			}
		}
	return result;
	}

#endif /* USE_OLD_PARSER */

#ifndef USE_OLD_PARSER

/* Parsing routing
 * Ryan Pierce (rdpierce@pobox.com)
 *
 * define this to get debugging output on the parse
 */
/*
#define DEBUG
*/
struct set *satisfy(struct formula *f) {
  
  struct formula *ptr; /* Pointer to current spot in formula */
  char szIndex[32]; /* Place to store the value of the index characters */
  int iIndex = 0; /* Index into the index char array */
  struct set *pResult; /* Return value */
  struct formula *pStartArg1 = NULL; /* Holds start of a new arg */
  struct formula *pEndArg1 = NULL; /* Holds last position of new arg */
  struct formula *pSaveArg1 = NULL; /* Holds where the next arg should be */
  struct formula *pStartArg2 = NULL;
  struct formula *pEndArg2 = NULL;
  struct formula *pSaveArg2 = NULL;
  int iParens; /* Number of parenthesis levels */
   
  /* Assume that all non-formula characters have been removed. Now we
   * must have either an index alone, or a ( followed by something inside
   * it. */
  
  ptr = f;
  if ( isdigit( ptr->character ) ) {
    
    /* It is a digit. The rest of the formula must be ONLY digits. */
#ifdef DEBUG
    printf( "Parsing index: " );
#endif
    do {
      if ( ! isdigit( ptr->character ) ) {
	printf( "satisfy: encountered symbol when parsing index\n" );
	exit( 1 );
      }
      szIndex[ iIndex++ ] = ptr->character;
#ifdef DEBUG
      printf( "%c", ptr->character );
#endif
      ptr = ptr->next;
    } while ( ptr != NULL );
    szIndex[ iIndex ] = NULL;
    pResult = ( struct set * ) malloc( sizeof( struct set ) );
    if ( pResult == NULL ) {
      printf( "satisfy: out of RAM\n" );
      exit( 1 );
    }
#ifdef DEBUG
    printf( " *DONE*\n" );
#endif
    pResult->index = atoi( szIndex );
    pResult->flags = SATISFY_TRUE;
    pResult->true = NULL;
    pResult->false = NULL;
    return pResult;

  } else if ( ptr->character == '(' ) {
    /* This is an expression */
    /* Now determine if this is going to be an and; it will be binary
     * with two formulas at this level with an ^ in the middle. Or it
     * may be unary; in which case recurse to strip off the outer parens
     */
    pStartArg1 = ptr->next; /* We're before the start of first sub-formula */
    iParens = 0;
#ifdef DEBUG
    printf( "Parsing: " );
#endif
    do {
      if ( ptr->next == NULL ) {
	printf( "satisfy: premature end of formula\n" );
	exit( 1 );
      }
      /* Tally the parens */
      if ( ptr->next->character == '(' ) 
	iParens++;
      else if ( ptr->next->character == ')' )
	iParens--;
      /* Now check to see if, at the net 0 paren level, there is an ^ */
      if ( ( iParens == 0 ) && ( ptr->next->character == '^' ) ) {
	/* YES - this is an and */
#ifdef DEBUG
	printf( " *end of AND arg 1* " );
#endif
	pEndArg1 = ptr;
	pSaveArg1 = pEndArg1->next;
	pEndArg1->next = NULL;
	pStartArg2 = pSaveArg1->next;
	ptr = pSaveArg1; /* We're before the start of second sub-formula */
	iParens = 0;
	do {
	  if ( ptr->next == NULL ) {
	    printf( "satisfy: premature end of formula\n" );
	    exit( 1 );
	  }
	  /* Tally the parens */
	  if ( ptr->next->character == '(' )
	    iParens++;
	  else if ( ptr->next->character == ')' )
	    iParens--;
	  /* Now look for an end to this! */
	  if ( iParens == -1 ) {
	    /* We've got the end to second argument */
	    /* First check that the closing paren finishes us off */
	    if ( ptr->next->next != NULL ) {
	      printf( "satisfy: garbage after end of binary expression\n" );
	      exit( 0 );
	    }
#ifdef DEBUG
	    printf( " *and DONE*\n" );
#endif
	    pEndArg2 = ptr;
	    pSaveArg2 = pEndArg2->next;
	    pEndArg2->next = NULL;
	    pResult = intersection( satisfy( pStartArg1 ), 
				    satisfy( pStartArg2 ) );
	    pEndArg1->next = pSaveArg1;
	    pEndArg2->next = pSaveArg2;
	    return pResult;
	  }
#ifdef DEBUG
	  printf( "%c", ptr->next->character );
#endif
	  ptr = ptr->next;
	} while ( 1 );
      }
      /* Now check for iParens == -1, meaning we are at the close of a unary */
      if ( iParens == -1 ) {
	/* unary formula. Recurse. */
	/* First check that the closing paren finishes us off */
	if ( ptr->next->next != NULL ) {
	  printf( "satisfy: garbage after end of unary expression\n" );
	  exit( 0 );
	}
#ifdef DEBUG
	printf( " *unary formula DONE*\n" );
#endif
	pEndArg1 = ptr;
	pSaveArg1 = ptr->next;
	pEndArg1->next = NULL;
	pResult = satisfy( pStartArg1 );
	pEndArg1->next = pSaveArg1;
	return pResult;
      }
#ifdef DEBUG
      printf( "%c", ptr->next->character );    
#endif
      ptr = ptr->next;
    } while ( 1 );
    
  } else if ( ptr->character == '!' ) {
    /* This is a complement. Note that it must end in a NULL hence no
     * need to count parens, just pass the entire thing to parser. */
#ifdef DEBUG
    printf( "Parsing Complement: \n" );
#endif
    if ( ptr->next == NULL ) {
      printf( "satisfy: premature end of complement\n" );
      exit( 1 );
    }
    return complement( satisfy( ptr->next ) );

  } else {
    /* This has absolutely no business being here */
    printf( "satisfy: Invalid start of formula %c\n", ptr->character );
    exit(1);
  }
  
  printf( "select: fatal programming error; unreachable code reached!" );
  exit(1);
  return NULL;
}

#endif /* ifndef USE_OLD_PARSER */
